/* Safe variant of kilo editor */
/* Based on the web-tutorial and contain many edits */

#include <fcntl.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <termios.h>
#include <errno.h>
#include <ctype.h>

#define CTRL_KEY(k) ((k) & 0x1f)
#define TAB_SIZE 8

struct termios orig_termios;
struct winsize ws;

char *scr_buf;
size_t scr_buf_size;

struct row {
	char *buf;
	size_t len;

	char *render;
	size_t rlen;
};

struct row *row_s;
size_t row_size;

/* Cursors */
unsigned int curx;
unsigned int cury;
unsigned int renx;

unsigned int rowoff;
unsigned int coloff;

char *file_path;

enum {
	FAILED_SAVE = 1,
	SUCCESS_SAVE,
	MARK
};
char status_type;
char modified_flag;

/* for mark */
char mark_flag;
unsigned int curx2;
unsigned int cury2;

/* Cleaners */
void bufFree(void) {
	if (scr_buf != NULL)
		free(scr_buf);

	scr_buf = NULL;
	scr_buf_size = 0;
}

void fileBufFree(void) {
	if (row_s != NULL) {
		if (row_size > 0)
			for (size_t i = 0; i < row_size; i++) {
				if (row_s[i].buf)
					free(row_s[i].buf);

				if (row_s[i].render)
					free(row_s[i].render);
			}

		free(row_s);
	}

	row_s = NULL;
	row_size = 0;
}

void die(int ret, char *msg) {
	write(STDOUT_FILENO, "\033[H\033[J", 6);

	bufFree();
	fileBufFree();

	if (msg != NULL)
		fprintf(stderr, "que: %s: %s\n", msg, strerror(errno));

	exit(ret);
}

/* Save */
char *buf2str(size_t *len) {
	size_t totalen = 0;
	for (size_t i = 0; i < row_size; i++)
		totalen += row_s[i].len + 1;

	*len = totalen;
	char *buf = malloc(totalen + 1);
	if (buf == NULL)
		die(1, "malloc in buf2str: Fatal error");

	char *p = buf;
	for (size_t i = 0; i < row_size; i++) {
		memcpy(p, row_s[i].buf, row_s[i].len);

		p += row_s[i].len;
		*p = '\n';
		p++;
	}

	*p = '\0';
	return buf;
}

void save(void) {
	size_t len = 0;
	char *buf = buf2str(&len);

	int fd = open(file_path, O_RDWR, 0644);
	if (fd < 0)
		goto BAD_SAVE;

	if (ftruncate(fd, len) < 0)
		goto BAD_SAVE;

	ssize_t ret = write(fd, buf, len);
	if (ret == -1)
		goto BAD_SAVE;

	else if ((size_t)ret != len)
		goto BAD_SAVE;

	close(fd);
	free(buf);
	status_type = SUCCESS_SAVE;
	modified_flag = 0;
	return;

BAD_SAVE:
	free(buf);
	status_type = FAILED_SAVE;
	return;
}

/* Funcs */
void bufAppend(const char *str, const size_t len) {
	if (len == 0)
		return;

	char *buf = NULL;
	if (scr_buf_size == 0 || scr_buf == NULL)
		buf = malloc(len);

	else
		buf = realloc(scr_buf, scr_buf_size + len);

	if (buf == NULL) {
		if (scr_buf)
			free(scr_buf);

		scr_buf_size = 0;
		return;
	}

	scr_buf = buf;

	memcpy(&scr_buf[scr_buf_size], str, len);
	scr_buf_size += len;
}

int updateRenderStr(const unsigned int row) {
	int tabs = 0;
	for (size_t i = 0; i < row_s[row].len; i++)
		if (row_s[row].buf[i] == '\t')
			tabs++;

	if (row_s[row].render)
		free(row_s[row].render);

	char *buf = malloc(row_s[row].len + tabs * TAB_SIZE + 1);
	if (buf == NULL)
		return 1;
	row_s[row].render = buf;

	int j = 0;
	for (size_t i = 0; i < row_s[row].len; i++) {
		if (row_s[row].buf[i] == '\t') {
			row_s[row].render[j++] = ' ';
			while (j % TAB_SIZE != 0)
				row_s[row].render[j++] = ' ';
		}

		else
			row_s[row].render[j++] = row_s[row].buf[i];
	}

	row_s[row].render[j] = '\0';
	row_s[row].rlen = j;

	return 0;
}

int addRow(const unsigned int at, const char *buf, const size_t len) {
	struct row *row_s2 = realloc(row_s, sizeof(struct row) * (row_size + 1));
	if (row_s2 == NULL)
		return 1;
	row_s = row_s2;

	memmove(&row_s[at + 1], &row_s[at], sizeof(struct row) * (row_size - at));

	char *buf2 = malloc(len + 1);
	if (buf2 == NULL)
		return 1;
	row_s[at].buf = buf2;

	memcpy(row_s[at].buf, buf, len);
	row_s[at].len = len;
	row_s[at].buf[len] = '\0';

	row_s[at].render = NULL;
	if (updateRenderStr(at))
		return 1;

	row_size++;
	return 0;
}

void readFile(const char *path) {
	row_s = malloc(sizeof(struct row));
	if (row_s == NULL)
		die(1, "malloc in readFIle");

	FILE *fp = fopen(path, "r");
	if (fp == NULL)
		die(1, "fopen in readFile");

	char *buf = NULL;
	size_t n = 0;
	ssize_t len = 0;
	while ((len = getline(&buf, &n, fp)) != -1) {
		while (len > 0 && (buf[len - 1] == '\n' || buf[len - 1] == '\r'))
			len--;

		if (addRow(row_size, buf, len)) {
			fclose(fp);
			free(buf);
			die(1, "addRow in readFile");
		}
	}

	if (buf)
		free(buf);

	fclose(fp);
}

void insertChar(char c) {
	if ((size_t)cury == row_size) {
		if (addRow(row_size, "", 0)) {
			save();
			die(1, "addRow in insertChar: Fatal error");
		}
	}

	size_t i = (size_t)curx;
	if (i > row_s[cury].len)
		i = row_s[cury].len;

	char *buf = realloc(row_s[cury].buf, row_s[cury].len + 2);
	if (buf == NULL) {
		save();
		die(1, "realloc in insertChar: Fatal error");
	}
	row_s[cury].buf = buf;

	memmove(&row_s[cury].buf[i + 1], &row_s[cury].buf[i], row_s[cury].len - i + 1);

	row_s[cury].len++;
	row_s[cury].buf[i] = c;
	if (updateRenderStr(cury)) {
		save();
		die(1, "updateRenderStr in insertChar: Fatal error");
	}

	curx++;
}

void insertNewline(void) {
	if (curx == 0) {
		if (addRow(cury, "", 0)) {
			save();
			die(1, "addRow in insertNewline: Fatal error");
		}
	}

	else {
		if (addRow(cury + 1, &row_s[cury].buf[curx], row_s[cury].len - curx)) {
			save();
			die(1, "addRow in insertNewline: Fatal error");
		}

		/* Resize */
		row_s[cury].len = (size_t)curx;
		row_s[cury].buf[curx] = '\0';

		char *buf = strdup(row_s[cury].buf);
		if (buf == NULL) {
			save();
			die(1, "strdup in insertNewline: Fatal error");
		}

		free(row_s[cury].buf);
		row_s[cury].buf = buf;

		if (updateRenderStr(cury)) {
			save();
			die(1, "updateRenderStr in insertNewline: Fatal error");
		}
	}

	cury++;
	curx = 0;
}

void delChar(void) {
	if ((curx == 0 && cury == 0) || cury == (unsigned int)row_size)
		return;

	else if (curx > 0) {
		if (curx - 1 >= (unsigned int)row_s[cury].len)
			return;

		memmove(&row_s[cury].buf[curx - 1], &row_s[cury].buf[curx], row_s[cury].len - curx);

		row_s[cury].len--;
		curx--;
	}

	else {
		curx = (unsigned int)row_s[cury - 1].len;

		char *buf = realloc(row_s[cury - 1].buf, row_s[cury - 1].len + row_s[cury].len + 1);
		if (buf == NULL) {
			save();
			die(1, "realloc in delChar: Fatal error");
		}
		row_s[cury - 1].buf = buf;

		memcpy(&row_s[cury - 1].buf[row_s[cury - 1].len], row_s[cury].buf, row_s[cury].len);
		row_s[cury - 1].len += row_s[cury].len;
		row_s[cury - 1].buf[row_s[cury - 1].len] = '\0';

		if (cury < (unsigned int)row_size) {
			free(row_s[cury].buf);
			free(row_s[cury].render);

			memmove(&row_s[cury], &row_s[cury + 1], sizeof(struct row) * (row_size - cury - 1));
			struct row *buf2 = realloc(row_s, sizeof(struct row) * row_size - 1);
			if (buf == NULL) {
				save();
				die(1, "relloc in delChar: Fatal error");
			}

			row_s = buf2;
			row_size--;
		}

		cury--;
	}

	if (updateRenderStr(cury)) {
		save();
		die(1, "updateRenderStr in delChar: Fatal error");
	}
}

void cmd(void) {
	/* Execute internal cmds */
}

/* Terminal */
void DRawMode(void) {
	if (tcsetattr(STDIN_FILENO, TCSAFLUSH, &orig_termios) < 0)
	die(1, "tcsetattr");
}

void ERawMode(void) {
	if (tcgetattr(STDIN_FILENO, &orig_termios) < 0)
		die(1, "tcgetattr");

	struct termios raw = orig_termios;
	raw.c_oflag &= ~(OPOST);
	raw.c_cflag |= (CS8);
	raw.c_iflag &= ~(BRKINT | ICRNL | INPCK | ISTRIP | IXON);
	raw.c_lflag &= ~(ECHO | ICANON | ISIG | IEXTEN);

	raw.c_cc[VMIN] = 0;
	raw.c_cc[VTIME] = 1;

	if (tcsetattr(STDIN_FILENO, TCSAFLUSH, &raw) < 0)
		die(1, "tcsetattr");
}

void winsize(void) {
	if (ioctl(STDOUT_FILENO, TIOCGWINSZ, &ws) < 0)
		die(1, "ioctl");

	if (ws.ws_col == 0 || ws.ws_row <= 1)
		die(1, "winsize");

	ws.ws_row--;
}

void statusBar(const char *path) {
	char info[124];
	int ret = 0;

	if (status_type == 0)
		ret = snprintf(info, sizeof(info), "LINE: %u/%zu %s [%s]", cury + 1, row_size + 1, (modified_flag) ? "[modified]" : "", path);

	else if (status_type == FAILED_SAVE)
		ret = snprintf(info, sizeof(info), "Failed save to %s: %s", path, strerror(errno));

	else if (status_type == SUCCESS_SAVE)
		ret = snprintf(info, sizeof(info), "File %s success saved", path);

	else if (status_type == MARK)
		ret = snprintf(info, sizeof(info), "Mark set at %ux %uy", curx, cury);

	status_type = 0;
	bufAppend(info, ret);
}

/* Keyboard */
char readkey(void) {
	char c;
	int ret;

	while ((ret = read(STDIN_FILENO, &c, 1)) != 1) {
		if (ret < 0 && errno != EAGAIN) {
			save();
			die(1, "read");
		}
	}

	return c;
}

void moveCursor(char c) {
	switch (c) {
		case 'A':
			if (cury != 0)
				cury--;

			break;

		case 'B':
			if (cury < row_size)
				cury++;

			break;

		case 'C':
			if (curx < row_s[cury].len)
				curx++;

			else if (curx == row_s[cury].len) {
				cury++;
				curx = 0;
			}

			break;

		case 'D':
			if (curx != 0)
				curx--;


			else if (cury > 0) {
				cury--;
				curx = row_s[cury].len;
			}

			break;

		case '5':
			if (readkey() == '~') {
				cury = rowoff;

				size_t t = ws.ws_row;
				while (t--)
					if (cury != 0)
						moveCursor('A');
			}

			break;

		case '6':
			if (readkey() == '~') {
				cury = rowoff + ws.ws_row + 1;

				if (cury > (unsigned int)row_size)
					cury = row_size;

				size_t t = ws.ws_row;
				while (t--)
					moveCursor('B');
			}

			break;

		case 'H':
			curx = 0;
			break;

		case 'F':
			curx = row_s[cury].len;
			break;
	}

	if (curx > row_s[cury].len)
		curx = row_s[cury].len;
}

void keyboard(void) {
	char key = readkey();
	switch (key) {
		case '\033':
			if (readkey() != '[')
				break;

			moveCursor(readkey());
			break;

		case '\r':
			modified_flag = 1;
			insertNewline();
			break;

		case CTRL_KEY('f'):
			cmd();
			break;

		case CTRL_KEY('q'):
			die(0, NULL);
			break;

		case CTRL_KEY('s'):
			save();
			break;

		case 127:
		case CTRL_KEY('h'):
			delChar();
			modified_flag = 1;
			break;

		default:
			modified_flag = 1;
			insertChar(key);
			break;
	}
}

int main(int argc, char **argv) {
	if (argc < 2) {
		printf("que [file]\n");
		return 0;
	}

	file_path = argv[1];
	readFile(file_path);
	winsize();
	atexit(DRawMode);
	ERawMode();

	while (1) {
		bufAppend("\033[H\033[J", 6);

		renx = 0;

		if (cury < row_size) {
			/* Buffer coord to Render buffer coord */
			for (unsigned int i = 0; i < curx; i++) {
				if (row_s[cury].buf[i] == '\t')
					renx += (TAB_SIZE - 1) - (renx % TAB_SIZE);

				renx++;
			}
		}

		if (cury < rowoff)
			rowoff = cury;

		else if (cury >= rowoff + ws.ws_row)
			rowoff = cury - ws.ws_row + 1;

		if (renx < coloff)
			coloff = renx;

		else if (renx >= coloff + ws.ws_col)
			coloff = renx - ws.ws_col + 1;

		for (size_t i = 0; i < ws.ws_row; i++) {
			unsigned int row = rowoff + i;

			if (row_size > row) {
				int len = row_s[row].rlen - coloff;
				if (len < 0)
					len = 0;

				else if (len > ws.ws_col)
					len = ws.ws_col;

				bufAppend(&row_s[row].render[coloff], (size_t)len);
			}

			bufAppend("\r\n", 2);
		}

		statusBar(file_path);

		char buf[32];
		int ret = snprintf(buf, sizeof(buf), "\033[%u;%uH", cury - rowoff + 1, renx - coloff + 1);
		bufAppend(buf, ret);

		write(STDOUT_FILENO, scr_buf, scr_buf_size);
		bufFree();

		keyboard();
	}
}
